@@ -119,7 +119,7 @@ struct MHD_NonceNc
    * Nonce counter, a value that increases for each subsequent
    * request for the same nonce.
    */
-  unsigned int nc;
+  unsigned long int nc;
 
   /**
    * Nonce value: 
@@ -379,11 +379,15 @@ enum MHD_CONNECTION_STATE
   MHD_CONNECTION_FOOTERS_SENT = MHD_CONNECTION_FOOTERS_SENDING + 1,
 
   /**
-   * 19: This connection is closed (no more activity
-   * allowed).
+   * 19: This connection is to be closed.
    */
   MHD_CONNECTION_CLOSED = MHD_CONNECTION_FOOTERS_SENT + 1,
 
+  /**
+   * 20: This connection is finished (only to be freed)
+   */
+  MHD_CONNECTION_IN_CLEANUP = MHD_CONNECTION_CLOSED + 1,
+
   /*
    *  SSL/TLS connection states
    */
@@ -440,10 +444,15 @@ struct MHD_Connection
 {
 
   /**
-   * This is a linked list.
+   * This is a doubly-linked list.
    */
   struct MHD_Connection *next;
 
+  /**
+   * This is a doubly-linked list.
+   */
+  struct MHD_Connection *prev;
+
   /**
    * Reference to the MHD_Daemon struct.
    */
@@ -598,6 +607,12 @@ struct MHD_Connection
    */
   time_t last_activity;
 
+  /**
+   * After how many seconds of inactivity should
+   * this connection time out?  Zero for no timeout.
+   */
+  unsigned int connection_timeout;
+
   /**
    * Did we ever call the "default_handler" on this connection?
    * (this flag will determine if we call the 'notify_completed'
@@ -621,6 +636,11 @@ struct MHD_Connection
    */
   int read_closed;
 
+  /**
+   * Set to MHD_YES if the thread has been joined.
+   */
+  int thread_joined;
+
   /**
    * State in the FSM for this connection.
    */
@@ -641,11 +661,6 @@ struct MHD_Connection
    */
   int response_unready;
 
-  /**
-   * Are we sending with chunked encoding?
-   */
-  int have_chunked_response;
-
   /**
    * Are we receiving with chunked encoding?  This will be set to
    * MHD_YES after we parse the headers and are processing the body
@@ -752,9 +767,24 @@ struct MHD_Daemon
   void *default_handler_cls;
 
   /**
-   * Linked list of our current connections.
+   * Tail of doubly-linked list of our current, active connections.
    */
-  struct MHD_Connection *connections;
+  struct MHD_Connection *connections_head;
+
+  /**
+   * Tail of doubly-linked list of our current, active connections.
+   */
+  struct MHD_Connection *connections_tail;
+
+  /**
+   * Tail of doubly-linked list of connections to clean up.
+   */
+  struct MHD_Connection *cleanup_head;
+
+  /**
+   * Tail of doubly-linked list of connections to clean up.
+   */
+  struct MHD_Connection *cleanup_tail;
 
   /**
    * Function to call to check if we should
@@ -852,15 +882,27 @@ struct MHD_Daemon
   pthread_t pid;
 
   /**
-   * Mutex for per-IP connection counts
+   * Mutex for per-IP connection counts.
    */
   pthread_mutex_t per_ip_connection_mutex;
 
+  /**
+   * Mutex for (modifying) access to the "cleanup" connection DLL.
+   */
+  pthread_mutex_t cleanup_connection_mutex;
+
   /**
    * Listen socket.
    */
   int socket_fd;
 
+#ifndef HAVE_LISTEN_SHUTDOWN
+  /**
+   * Pipe we use to signal shutdown.
+   */
+  int wpipe[2];
+#endif
+
   /**
    * Are we shutting down?
    */
@@ -971,5 +1013,44 @@ struct MHD_Daemon
 #endif
 
 
+/**
+ * Insert an element at the head of a DLL. Assumes that head, tail and
+ * element are structs with prev and next fields.
+ *
+ * @param head pointer to the head of the DLL
+ * @param tail pointer to the tail of the DLL
+ * @param element element to insert
+ */
+#define DLL_insert(head,tail,element) do { \
+  (element)->next = (head); \
+  (element)->prev = NULL; \
+  if ((tail) == NULL) \
+    (tail) = element; \
+  else \
+    (head)->prev = element; \
+  (head) = (element); } while (0)
+
+
+/**
+ * Remove an element from a DLL. Assumes
+ * that head, tail and element are structs
+ * with prev and next fields.
+ *
+ * @param head pointer to the head of the DLL
+ * @param tail pointer to the tail of the DLL
+ * @param element element to remove
+ */
+#define DLL_remove(head,tail,element) do { \
+  if ((element)->prev == NULL) \
+    (head) = (element)->next;  \
+  else \
+    (element)->prev->next = (element)->next; \
+  if ((element)->next == NULL) \
+    (tail) = (element)->prev;  \
+  else \
+    (element)->next->prev = (element)->prev; \
+  (element)->next = NULL; \
+  (element)->prev = NULL; } while (0)
+
 
 #endif
